//библиотеки I2Cdev, MPU6050, Adafruit Unified Sensor, Adafruit BME280 должны быть установлены,
//или файлы .cpp/.h этих классов помещены в каталог библиотек Arduino
#include "I2Cdev.h"
//библиотека Arduino Wire требуется для имплементации
//I2CDEV_ARDUINO_WIRE в I2Cdev.h
#if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
    #include "Wire.h"
#endif

//интерфейс I2C: A4-SDA, A5-SCL
#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>

#include "MPU6050_6Axis_MotionApps20.h"
//#include "MPU6050.h" // не требуется при использовании MotionApps

#define a_GAIN0 6.8 //опорный порог замедления для сухого покрытия (ф=0.8), м/с2. В формате int или float
#define a_GAIN1 4.3 //порог замедления для влажного покрытия (ф=0.5), м/с2. В формате int или float
#define G 9.8 //g=9.8 м/с2
#define ang_MAX 12 //пороговый угол тангажа/крена, град. В формате int

#define T_LED 150 //интервал вспышек, мс
#define T_ESS 2000 //длительность сигнала ESS, мс
#define T_MOT 1000 //длительность сигнала детектора движения, мс
#define T_MES 10000 //период измерения влажности и температуры, мс

#define INT_PIN 2 //вход для имп. прерываний с вых. INT MPU6050 
#define ESS_PIN 13 //основной выход ESS
#define ESS2_PIN 12 //доп. выход ESS
#define MOT_PIN 10 //выход детектора движения (DMP)
#define IND_PIN 11 //выход индикатора тангажа/крена и ESS
//активные уровни сигналов
#define ESS_ACT_OUT LOW
#define ESS2_ACT_OUT HIGH
#define MOT_ACT_OUT HIGH

Adafruit_BME280 bme;
//I2C адрес BME280 по умолчанию 0x76
MPU6050 accelgyro;
//I2C адрес MPU6050 по умолчанию 0x68
//AD0 low = 0x68
//AD0 high = 0x69
//MPU6050 mpu(0x69); // использовать для AD0 high

//переменные контроля/статуса MPU
bool dmpReady = false,  //флаг готовности DMP
     bmeReady = false;  //флаг готовности BME
uint8_t intStatus;      //байт статуса регистра прерываний MPU
uint8_t devStatus;      //статус операции (0 = success, !0 = error)
uint16_t packetSize;    //размер пакета данных DMP (по умолчанию 42 байта)
uint16_t fifoCount;     //текущее кол-во байт в FIFO
uint8_t fifoBuffer[64]; //буфер хранения FIFO

//переменные ориентации/движения
Quaternion q;           //[w, x, y, z]         контейнер кватерниона
VectorInt16 aa;         //[x, y, z]            данные акселератора
VectorInt16 aaReal;     //[x, y, z]            данные акселератора за вычетом g
VectorFloat gravity;    //[x, y, z]            вектор гравитации

volatile bool mpuInt = false;     //флаг прерывания MPU
//обработчик прерывания
void mpuInterrupt() {
  mpuInt = true;
}

float ypr[3], //[yaw, pitch, roll]   контейнер данных в системе YPR
      rH, //влажность, %
      T;  //температура, С

int aY0, aY1, aY_max, //диапазон и предельное значение замедления
    aP, aR; //углы крена/тангажа
unsigned int n, //расчетное количество вспышек за T_ESS
             k, i; //переменные числа циклов
boolean ess, ang, motion; //флаги ESS, крена/тангажа и детектора движения
unsigned long t_led_, t_mot_, t_mes_, t_; //переменные для расчета таймингов

void setup() {
    Serial.begin(115200);
    //назначение входов/выходов
    pinMode(INT_PIN, INPUT_PULLUP);
    pinMode(ESS_PIN, OUTPUT);
    pinMode(ESS2_PIN, OUTPUT);
    pinMode(MOT_PIN, OUTPUT);
    pinMode(IND_PIN, OUTPUT);
    digitalWrite(ESS_PIN, !ESS_ACT_OUT);
    digitalWrite(ESS2_PIN, !ESS2_ACT_OUT);
    digitalWrite(MOT_PIN, !MOT_ACT_OUT);
    digitalWrite(IND_PIN, LOW);
    
    //расчет значений замедления
    aY0 = round(16384*(float)a_GAIN0/G);
    aY1 = round(16384*(float)a_GAIN1/G);
    Serial.println("aY0: "+String(aY0));
    Serial.println("aY1: "+String(aY1));
    n = round(T_ESS/T_LED);
    Serial.println("n: "+String(n));

    //настройка I2C
    #if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
        Wire.begin();
        TWBR = 24; //частота шины I2C 400kHz (200kHz для CPU 8MHz)
    #elif I2CDEV_IMPLEMENTATION == I2CDEV_BUILTIN_FASTWIRE
        Fastwire::setup(400, true);
    #endif

    //инициализация BME280
    Serial.println(F("Initializing BME..."));
    if (bme.begin(0x76)) bmeReady = true;                                
      else Serial.println("Could not find a BME280 sensor");
    
    //инициализация MPU6050
    accelgyro.initialize();
    
    //загрузка и настройка DMP
    Serial.println(F("Initializing DMP..."));
    devStatus = accelgyro.dmpInitialize();
    
    /* FS_SEL определяет диапазон акселерометра:
     * 0 = +/- 2g (по умолчанию)
     * 1 = +/- 4g
     * 2 = +/- 8g
     * 3 = +/- 16g
    */
    accelgyro.setFullScaleAccelRange(0);
    
    //настройки чувствительности детектора движения
    /* ACCEL_HPF
    * 0         | Reset       | None
    * 1         | On          | 5Hz
    * 2         | On          | 2.5Hz
    * 3         | On          | 1.25Hz
    * 4         | On          | 0.63Hz
    * 7         | Hold        | None
    */
    accelgyro.setDHPFMode(3);
    /* Порог чувствительности детектора движения
    * LSB (ЦОД) = 2mg
    */
    accelgyro.setMotionDetectionThreshold(1);
    /** Длительность события детектора движения
    * LSB (ЦОД) = 1ms
    */
    accelgyro.setMotionDetectionDuration(5);

    //проверка статуса MPU
    if (devStatus == 0) {
       //включение DMP
       Serial.println(F("Enabling DMP..."));
       accelgyro.setDMPEnabled(true);
    
       //разрешение прерывания
       attachInterrupt(digitalPinToInterrupt(INT_PIN), mpuInterrupt, RISING);
       //attachInterrupt(0, mpuInterrupt, RISING); //также interrupt 0, PIN 2
            
       intStatus = accelgyro.getIntStatus();
    
       //установка флага готовности DMP
       Serial.println(F("DMP ready! Waiting for first interrupt..."));
       dmpReady = true;
    
       //получение размера пакета данных DMP
       packetSize = accelgyro.dmpGetFIFOPacketSize();
    } else {
       //ERROR!
       //1 = ошибка начальной загрузки
       //2 = ошибка конфигурации DMP
       Serial.print(F("DMP Initialization failed (code "));
       Serial.print(devStatus);
       Serial.println(F(")"));
    }
        
    t_mot_ = t_led_ = t_mes_ = t_ = millis();    
}

void loop() {
    if (!dmpReady || !bmeReady) return;        
    
    while(!mpuInt && fifoCount < packetSize){
      //пустой цикл до момента прерывания по INT_PIN         
    }
    //выход из цикла по прерыванию    
    //сброс флага прерывания и чтение байта INT_STATUS
    mpuInt = false;    
    intStatus = accelgyro.getIntStatus();

    //чтение счетчика FIFO
    fifoCount = accelgyro.getFIFOCount();

    //проверка переполнения
    if ((intStatus & 0x10) || fifoCount == 1024) {
      //очистка FIFO
      accelgyro.resetFIFO();
      Serial.println(F("FIFO overflow!"));

    //проверка регистра прерываний DMP
    } else if (intStatus & 0x01) {
        //ожидание пакета данных нужного размера
        while (fifoCount < packetSize) fifoCount = accelgyro.getFIFOCount();

        //чтение из FIFO
        accelgyro.getFIFOBytes(fifoBuffer, packetSize);
        
        //проверка счетчика FIFO на наличие более одного пакета
        //(позволяет получить данные немедленно без ожидания прерывания)
        fifoCount -= packetSize;

        //получение величины реального ускорения
        //за вычетом гравитации
        accelgyro.dmpGetQuaternion(&q, fifoBuffer);
        accelgyro.dmpGetAccel(&aa, fifoBuffer);
        accelgyro.dmpGetGravity(&gravity, &q);
        accelgyro.dmpGetLinearAccel(&aaReal, &aa, &gravity);
        
        //Serial.print("aaReal\t");
        //Serial.print(aaReal.x);
        //Serial.print("\t");
        Serial.println(aaReal.y);
        //Serial.print("\t");
        //Serial.println(aaReal.z);

        accelgyro.dmpGetYawPitchRoll(ypr, &q, &gravity);

        aP = round(ypr[1] * 180/M_PI); //крен
        aR = round(ypr[2] * 180/M_PI); //тангаж
        
        //Serial.print("ypr\t");
        //Serial.print(ypr[0] * 180/M_PI);
        Serial.print("\t");
        Serial.print(aP); //угол вокруг оси x
        Serial.print("\t");
        Serial.println(aR); //угол вокруг оси y
        
        //превышение предела замедления
        if(abs(aaReal.y)>=aY_max){
          ess = true;
        }

        //превышение угла тангажа/крена
        //с приоритетом индикации ESS
          if(abs(aP)>=ang_MAX || abs(aR)>=ang_MAX){
            if(!ess) digitalWrite(IND_PIN, HIGH);
            ang = true;
            Serial.println("Max angle!");
          } else {
            if(!ess) digitalWrite(IND_PIN, LOW);
            ang = false;
          }        
    }

    //сигналы детектора движения по трем осям в разных направлениях
    if(accelgyro.getXNegMotionDetected() || accelgyro.getXPosMotionDetected() 
       || accelgyro.getYNegMotionDetected() || accelgyro.getYPosMotionDetected()
       || accelgyro.getZNegMotionDetected() || accelgyro.getZPosMotionDetected()){
       Serial.println("Motion Detected");
       motion = true;
    }

    unsigned long t = millis();

    //измерение влажности и температуры с периодом T_MES
     if(t_mes_ < t){
        t_mes_ = t + T_MES;
        
        rH = bme.readHumidity();
        T = bme.readTemperature();
        //вычисление порога замедления по измеренному значению влажности (%, float)
        if (T>0) aY_max = round((aY1-aY0)*rH/100.0 + aY0);
          else aY_max = aY1; //при T<0C
        Serial.println("rH: "+String(rH));
        Serial.println("TC: "+String(T));
        Serial.println("aY_max: "+String(aY_max));     
     }
    
    //подача сигнала ESS     
    if(ess){
      if(t_led_ < t){ //подача сигналов с периодом T_ESS 
            t_led_ = t + T_LED;
            digitalWrite(ESS_PIN, !digitalRead(ESS_PIN));
            digitalWrite(ESS2_PIN, !digitalRead(ESS2_PIN));
            digitalWrite(IND_PIN, !digitalRead(IND_PIN)); //индикатор
            i++;
            //Serial.println("i: "+String(i));
            if(i>n){ //прекратить подачу сигнала ESS
              digitalWrite(ESS_PIN, !ESS_ACT_OUT);
              digitalWrite(ESS2_PIN, !ESS2_ACT_OUT);
              digitalWrite(IND_PIN, LOW);
              ess = false;
              i = 0;
            }
          }
     }     

     //сигнал детектора движения    
     if(motion){
       if(t_mot_ < t){
          t_mot_ = t + T_MOT;
          digitalWrite(MOT_PIN, !digitalRead(MOT_PIN));
          //Serial.println("Motion detector signal");
          if(!digitalRead(MOT_PIN)){
            motion = false;
          }
        }         
     }
     
     //индикация готовности
     //краткие вспышки с периодом 10xT_LED
     if(!ess && !ang && digitalRead(ESS_PIN)!= ESS_ACT_OUT){
        if(t_ < t){
          t_ = t + T_LED;
          k++;
          if(k < 10){
            digitalWrite(IND_PIN, LOW);    
          } else {
            digitalWrite(IND_PIN, HIGH);
            k = 0;
          }
        }
     }
}
